"""
a conditional generative adversarial network designed specifically for working
with cryptopunks art projects
"""

class ConditionalGAN(Model):
    def __init__(self):
        self.batch_size = 64
        self.num_channels = 1
        self.num_classes = 10
        self.image_size = 28
        self.latent_dim = 128
        self.learning_rate = 0.0002
        self.beta_1 = 0.5
        self.generator_in_channels = self.latent_dim + self.num_classes
        self.discriminator_in_channels = self.num_channels + self.num_classes
        self.opt = Adam(lr=self.learning_rate, 
                        beta_1 = self.beta_1)
        
    def __setattr__(self, name, value):
        print(name)
        return super().__setattr__(name, value)
    
    # define the standalone discriminator model
    def define_discriminator(self, in_shape=(24, 24, 3)):
        self.discriminator = Sequential()
        # downsample
        self.discriminator.add(Conv2D(128, (3,3), strides=(2,2), padding='same', input_shape=in_shape))
        self.discriminator.add(LeakyReLU(alpha=0.2))
        # downsample
        self.discriminator.add(Conv2D(128, (3,3), strides=(2,2), padding='same'))
        self.discriminator.add(LeakyReLU(alpha=0.2))
        # classifier
        self.discriminator.add(Flatten())
        self.discriminator.add(Dropout(0.4))
        self.discriminator.add(Dense(1, activation='sigmoid'))
        
    def define_generator(self):
        self.generator = Sequential()
        # foundation for 6*6 image
        n_nodes = 128 * 6 * 6
        self.generator.add(Dense(n_nodes, 
                                 input_dim=self.latent_dim))
        self.generator.add(LeakyReLU(alpha=0.2))
        self.generator.add(Reshape((6, 6, 128)))
        # upsample to 12x12
        self.generator.add(Conv2DTranspose(128, (4,4), strides=(2,2), padding='same'))
        self.generator.add(LeakyReLU(alpha=0.2))
        # upsample to 24x24
        self.generator.add(Conv2DTranspose(128, (4,4), strides=(2,2), padding='same'))
        self.generator.add(LeakyReLU(alpha=0.2))
        # generate
        self.generator.add(Conv2D(3, (6, 6), activation='tanh', padding='same'))

        
    def define_gan(generator, discriminator):
        # make weights in the discriminator not trainable
        discriminator.trainable = False
        # connect them
        self.model = Sequential()
        # add generator
        self.model.add(self.generator)
        # add the discriminator
        self.model.add(discriminator)
        
    def generate_real_samples(self):
        # choose random instances
        ix = randint(0, self.dataset.shape[0], self.n_samples)
        # select images
        X = self.dataset[ix]
        # generate class labels
        y = ones((self.n_samples, 1))
        return X, y
    
    def generate_latent_points(self):
        'generate points in latent space as input for the generator'
        # generate points in the latent space
        x_input = randn(latent_dim * n_samples)
        # reshape into a batch of inputs for the network
        x_input = x_input.reshape(n_samples, latent_dim)
        return x_input
    
    def generate_fake_samples(self):
        # generate points in latent space
        x_input = generate_latent_points(latent_dim, n_samples)
        # predict outputs
        X = generator.predict(x_input)
        # create class labels
        y = zeros((n_samples, 1))
        return X, y

    def train(n_epochs=1):
        '''train the generator and discriminator'''
        bat_per_epo = int(dataset.shape[0] / n_batch)
        half_batch = int(n_batch / 2)
        # manually enumerate epochs
        for i in range(n_epochs):
            # enumerate batches over the training set
            for j in range(bat_per_epo):
                # get randomly selected 'real' samples
                X_real, y_real = generate_real_samples(dataset, half_batch)
                # update discriminator model weights
                d_loss1, _ = d_model.train_on_batch(X_real, y_real)
                # generate 'fake' examples
                X_fake, y_fake = generate_fake_samples(g_model, latent_dim, half_batch)
                # update discriminator model weights
                d_loss2, _ = d_model.train_on_batch(X_fake, y_fake)
                # prepare points in latent space as input for the generator
                X_gan = generate_latent_points(latent_dim, n_batch)
                # create inverted labels for the fake samples
                y_gan = ones((n_batch, 1))
                # update the generator via the discriminator's error
                g_loss = gan_model.train_on_batch(X_gan, y_gan)
                # summarize loss on this batch
                print('>%d, %d/%d, d1=%.3f, d2=%.3f g=%.3f' %
                    (i+1, j+1, bat_per_epo, d_loss1, d_loss2, g_loss))
        # save the generator model
        g_model.save('generator.h5')
        
    def compile(self, opt):
        # compile model
        opt = Adam(lr=0.0002, 
                   beta_1=0.5)
        self.discriminator.compile(loss='binary_crossentropy', 
                                   optimizer=opt, 
                                   metrics=['accuracy'])
        self.model.compile(loss='binary_crossentropy', 
                           optimizer=opt)
        
    def predict(self):
     return True



from tensorflow.python.keras.callbacks import Callback, LearningRateScheduler
import numpy as np
import matplotlib.pyplot as plt
import os


# CALLBACKS #####
class CustomCallback(Callback):

    def __init__(self, run_folder, print_every_n_batches, initial_epoch, vae):
        self.epoch = initial_epoch
        self.run_folder = run_folder
        self.print_every_n_batches = print_every_n_batches
        self.vae = vae

    def on_train_batch_end(self, batch, logs={}):
        if batch % self.print_every_n_batches == 0:
            z_new = np.random.normal(size=(1, self.vae.z_dim))
            reconst = self.vae.decoder.predict(np.array(z_new))[0].squeeze()

            filepath = os.path.join(
                self.run_folder,
                'images',
                'img_' + str(self.epoch).zfill(3) + '_' + str(batch) + '.jpg'
            )

            if len(reconst.shape) == 2:
                plt.imsave(filepath, reconst, cmap='gray_r')
            else:
                plt.imsave(filepath, reconst)

    def on_epoch_begin(self, epoch, logs={}):
        self.epoch += 1


def step_decay_schedule(initial_lr, decay_factor=0.5, step_size=1):
    """
    Wrapper function to create a LearningRateScheduler with step decay
    schedule.
    """
    def schedule(epoch):
        new_lr = initial_lr * (decay_factor ** np.floor(epoch/step_size))

        return new_lr

    return LearningRateScheduler(schedule)



import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import ImageGrid


def show(X, idx=[3307, 281, 510, 741], ncols=2):
    Xs = [X[i] for i in idx]
    nrows = int(len(idx) / ncols)

    fig = plt.figure(figsize=(3., 3.))
    grid = ImageGrid(
        fig,
        111,  # similar to subplot(111)
        nrows_ncols=(nrows, ncols),  # creates 2x2 grid of axes
        axes_pad=0.1,  # pad between axes in inch.
    )

    for ax, im in zip(grid, Xs):
        ax.imshow(im)
    plt.show()



from tensorflow.python.keras.layers import (
    Input, Conv2D, Flatten, Dense, Conv2DTranspose, Reshape,
    Activation, BatchNormalization, LeakyReLU, Dropout
)
from tensorflow.python.keras.models import Model
from tensorflow.python.keras import backend as K
from tensorflow.python.keras.optimizers import Adam

from .callbacks import step_decay_schedule

import numpy as np


class Autoencoder():
    def __init__(
        self, input_dim, encoder_conv_filters, encoder_conv_kernel_size,
        encoder_conv_strides, decoder_conv_t_filters,
        decoder_conv_t_kernel_size, decoder_conv_t_strides, z_dim,
        use_batch_norm=False, use_dropout=False
    ):

        self.name = 'autoencoder'

        self.input_dim = input_dim
        self.encoder_conv_filters = encoder_conv_filters
        self.encoder_conv_kernel_size = encoder_conv_kernel_size
        self.encoder_conv_strides = encoder_conv_strides
        self.decoder_conv_t_filters = decoder_conv_t_filters
        self.decoder_conv_t_kernel_size = decoder_conv_t_kernel_size
        self.decoder_conv_t_strides = decoder_conv_t_strides
        self.z_dim = z_dim

        self.use_batch_norm = use_batch_norm
        self.use_dropout = use_dropout

        self.n_layers_encoder = len(encoder_conv_filters)
        self.n_layers_decoder = len(decoder_conv_t_filters)

        self._build()

    def _build(self):

        # THE ENCODER #####
        encoder_input = Input(shape=self.input_dim, name='encoder_input')

        x = encoder_input

        for i in range(self.n_layers_encoder):
            conv_layer = Conv2D(
                filters=self.encoder_conv_filters[i],
                kernel_size=self.encoder_conv_kernel_size[i],
                strides=self.encoder_conv_strides[i],
                padding='same',
                name='encoder_conv_' + str(i)
            )

            x = conv_layer(x)

            x = LeakyReLU()(x)

            if self.use_batch_norm:
                x = BatchNormalization()(x)

            if self.use_dropout:
                x = Dropout(rate=0.25)(x)

        shape_before_flattening = K.int_shape(x)[1:]

        x = Flatten()(x)
        encoder_output = Dense(self.z_dim, name='encoder_output')(x)

        self.encoder = Model(encoder_input, encoder_output)

        # THE DECODER #####
        decoder_input = Input(shape=(self.z_dim,), name='decoder_input')

        x = Dense(np.prod(shape_before_flattening))(decoder_input)
        x = Reshape(shape_before_flattening)(x)

        for i in range(self.n_layers_decoder):
            conv_t_layer = Conv2DTranspose(
                filters=self.decoder_conv_t_filters[i],
                kernel_size=self.decoder_conv_t_kernel_size[i],
                strides=self.decoder_conv_t_strides[i],
                padding='same',
                name='decoder_conv_t_' + str(i)
            )

            x = conv_t_layer(x)

            if i < self.n_layers_decoder - 1:
                x = LeakyReLU()(x)

                if self.use_batch_norm:
                    x = BatchNormalization()(x)

                if self.use_dropout:
                    x = Dropout(rate=0.25)(x)
            else:
                x = Activation('sigmoid')(x)

        decoder_output = x

        self.decoder = Model(decoder_input, decoder_output)

        # THE FULL AUTOENCODER #####
        model_input = encoder_input
        model_output = self.decoder(encoder_output)

        self.model = Model(model_input, model_output)

    def compile(self, learning_rate):
        self.learning_rate = learning_rate

        optimizer = Adam(lr=learning_rate)

        def r_loss(y_true, y_pred):
            return K.mean(K.square(y_true - y_pred), axis=[1, 2, 3])

        self.model.compile(optimizer=optimizer, loss=r_loss)

    def train(
        self, x_train, batch_size, epochs, print_every_n_batches=100,
        initial_epoch=0, lr_decay=1
    ):

        lr_sched = step_decay_schedule(
            initial_lr=self.learning_rate, decay_factor=lr_decay, step_size=1
        )

        callbacks_list = [lr_sched]

        self.model.fit(
            x_train, x_train, batch_size=batch_size, shuffle=True,
            epochs=epochs, initial_epoch=initial_epoch,
            callbacks=callbacks_list
        )



import numpy as np
import pandas as pd
import pathlib


class cpunks10k:
    def __init__(self):
        module_root = pathlib.Path(__file__).parents[0]
        self.punks_df = pd.read_pickle(f"{module_root}/data/punks.pkl")

    def any_to_one(self, i):
        x = True
        if i['any']:
            x = False
        return int(x)

    def load_data(self, labels=None):
        '''todo: reasonably handle labels or punt to another function,
                 flexible insample/outsample scheme,
                 think through insample/outsample to mix and insure
                 reasonable distribution of types.
        '''
        if labels is None:
            labels = list(self.punks_df.dtypes[0:-1].index)
        df = self.punks_df.copy()[labels + ['img']]
        df['any'] = df[labels].apply(np.any, axis=1)
        df['none'] = df.apply(lambda x: self.any_to_one(x), axis=1)

        df = df.drop(['any'], axis=1)
        X = df['img'].to_numpy()
        Y = df[labels + ['none']].to_numpy()
        X = np.array([row[0] for row in X])
        X_train = X[0:9000]
        Y_train = Y[0:9000]
        X_test = X[9000:10000]
        Y_test = Y[9000:10000]
        return (X_train, Y_train), (X_test, Y_test), (labels)



"""Tools for working with Cryptopunk NFTs; this includes utilities for data analysis and image preparation for training machine learning models using Cryptopunks as training data. 

Functions:
    get_punk(id)
    pixel_to_img(pixel_str, dim)
    flatten(img)
    unflatten(img)
    sort_dict_by_function_of_value(d, f)
    add_index_to_colors(colors)

"""
import os
import time
import requests
from collections import OrderedDict
from bs4 import BeautifulSoup
from re import sub
import numpy as np
import pandas as pd

from matplotlib.colors import rgb2hex
import matplotlib.image as mpimg
import matplotlib.pyplot as plt

__ROOT_DIR__ = os.path.dirname(os.path.abspath(__file__))
__PUNK_DIR__ = f"{__ROOT_DIR__}/images/training";

def camel_case(string):
  '''
     Convert string to camelCase
  '''
  string = string.strip("\n")
  string = sub(r"(_|-)+", " ", string).title().replace(" ", "")
  return string[0].lower() + string[1:]

def color_str_to_hex(s):
  '''
     Convert string representation of numpy pixel array 
     to a string hex value
  '''
  return rgb2hex([float(x) for x in s[1:-1].split(' ') if x != ''])

def get_punk(id):
    '''
       Returns a ndarray with loaded image
    ''' 
    return mpimg.imread(f'''{__PUNK_DIR__}/punk{"%04d" % id}.png''')

def pixel_to_img(pixel_str, dim = (24,24)):
    '''
       Take pixel of format "[r,g,b,b]"
       and return an image of size `dim` containing 
       only the pixel's color.
    '''
    (x,y) = dim 
    c = np.fromstring(pixel_str[1:-1], float, sep=' ')
    return np.full((x, y, 4), c)

def pixel_to_ximg(pixel_strs, dim = (24,24), n=3 ):
    '''
       Take pixel of format "[r,g,b,b]"
       and return an image of size `dim` containing
       a matrix of size n*n
    '''
    (x,y) = (dim[0]//n, dim[1]//n)
    m = []
    for i in range(0,n):
        l=[]
        for j in range(0,n):
            img = np.full((x, y, 4),
                           np.fromstring(pixel_strs[i*n + j][1:-1], float, sep=' '))
            l.append(img)
        m.append(np.concatenate(l, axis=1))
    return np.concatenate(m, axis=0)

def flatten(img): 
  '''
     Convert (x,y,z) array containing a pixel in z-dimension 
     to an (x,y) array with str values for each (i,j)
     the intention is to make this easier to work with in ML
     training.
  '''
  return np.array([[str(c) for c in row] 
                   for row in img])

def unflatten(img):
  '''
     Return a flattend image to valid .png format for display
  '''  
  return np.array([[np.fromstring(c[1:-1], float, sep=' ') 
                    for c in row] for row in img])

def sort_dict_by_function_of_value(d, f = len):
  sorted_tuples = sorted(d.items(),
          key=lambda item: len(item[1]))
  return {k: v for k, v in sorted_tuples}

def add_index_to_colors(colors):
  '''
     Add a unique, sequential index to the entry for 
     each color. returned dictionary will be of form
     {`color_string`: { `"id": `int`, "punkIds" : `list[int`}}
  '''
  i=0
  d={}
  for k in colors.keys():
    d[k] = {
      'id' : i,
      'punkIds' : colors[k]
    }
    i=i+1
  return d

def get_attr_dict():
    '''
       Read the attr csv and populate a default dict
    '''
    d=OrderedDict()
    with open(f"{__ROOT_DIR__}/data/list_attr_punx.csv") as f:
        for attr in f.read().strip('\n').split(','):
            d[attr]=-1
    return d

def get_punk_attrs(id):
    '''
       Retrieve `id` cryptopunk from larvalabs.com,
       parse HTML to extract type and attribute list
       to return list of attributes
    '''
    typeClass="col-md-10 col-md-offset-1 col-xs-12"
    punk_page=requests.get(f"https://www.larvalabs.com/cryptopunks/details/{id}")
    if(punk_page.status_code != 200):
        print(punk_page.status_code)
        return {}
    punk_html=punk_page.text

    soup = BeautifulSoup(punk_html, 'html.parser')
    details = soup.find(id="punkDetails")

    punkType = camel_case(details.find(class_=typeClass).find('a').contents[0])

    attrs=[punkType]
    attrTags = details.find(class_ = "row detail-row")
    for attrTag in attrTags.find_all('a'):
        attrs.append(camel_case(attrTag.contents[0]))
    return attrs


def get_punk_dict(id):
   '''
      Retrieve a punk page, pull type and attributes
      from HTML and return a dictionary of attribute to
      (-1,1) mapping where 1 is truthy for existence of 
      attribute
   '''
   od = {k:__ATTR_DICT__[k] for k in __ATTR_DICT__} 
   attrs = get_punk_attrs(id)
   for attr in attrs:
      od[attr]=1
   return od

def get_punks(start, end):
    '''
      Retrieve punks in range `start` to `end`
    '''
    punks={}
    for id in range(start, end):
        print(id)
        time.sleep(3.3)
        punks[id] = get_punk_dict(id)
    return punks

def plot_in_grid(n, images, predictions, labels):
  '''
    Plot `images` in an n*n grid with
    prediction and labels as header
  '''
  (x,y) = (n,n)
  fig = plt.figure(figsize=(9,14))
  i=0
  for i in range(1,(x*y)+1):
    fig.add_subplot(x, y, i)
    plt.imshow(images[i])
    plt.title(f"{predictions[i][0]},{labels[i][0]}")
    plt.axis('off')
    i=i+1
  return fig



from setuptools import setup, find_packages
import os
import fnmatch

from cpunks import __version__


DATA_DIRS = [
    'cpunks/data',
    'cpunks/tutorial',
]

DATA_EXTENSIONS = [
    '*.pickle',
    '*.pkl',
    '*.json',
    '*.ipynb'
]


def get_data_files():
    data_files = []
    for data_dir in DATA_DIRS:
        matches = []
        for file in os.listdir(data_dir):
            for exc_pattern in DATA_EXTENSIONS:
                if fnmatch.fnmatch(file, exc_pattern):
                    dst_path = os.path.join(data_dir, file)
                    matches.append(dst_path)
                    break
        data_files.append((data_dir, matches))
    return data_files


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='cpunks-10k',
    version=__version__,
    author='Tnn1t1s',
    author_email='tnn1t1s@protonmail.com',
    description='CPUNKS-10K are subsets of the 10,000 labeled images in the CryptoPunks collection, organized & modified for use in Machine Learning research',
    long_description=long_description,
    # long_description_cont_type="text/markdown",
    url='http://github.com/tnn1t1s/cpunks-10k',
    license='BSD',
    packages=find_packages(),
    include_package_data=True,
    data_files=get_data_files(),
    scripts=['bin/cpunk'],
    install_requires=[
        'numpy',
        'pandas',
        'matplotlib',
        'tensorflow',
        'jupyter',
    ],
    extras_require={
        'dev': [
            'flake8',
        ]
    },
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only', 
    ]
)



import os
import sys
import argparse
import shutil


# Top

cli_parser = argparse.ArgumentParser(os.path.basename(sys.argv[0]))

def usage(args):
    cli_parser.print_usage()

cli_parser.set_defaults(fn=usage)

cli_subparsers = cli_parser.add_subparsers(
    title='subcommands',
    description='valid subcommands',
    help='additional help'
)


# Tutorial

tutorial_parser = cli_subparsers.add_parser('tutorial')
tutorial_parser.add_argument('init', help='initial tutorial data')

def init_tutorial(args):
    dirname = os.path.dirname(os.path.abspath(__file__))
    tutorialdir = os.path.join(dirname, 'tutorial')
    shutil.copytree(tutorialdir, "./tutorial")
    datadir = os.path.join(dirname, 'data')
    shutil.copytree(datadir, "./data")

tutorial_parser.set_defaults(fn=init_tutorial)

